"""Client for interfacing with the thermostat"""

from __future__ import annotations

import asyncio

from collections.abc import Callable
from logging import Logger
from typing import Any

from .const import Action, FunctionalDomain
from .packet import decode_packet
from .socket_client import SocketClient
from .utils import encode_temperature, generate_command_bytes


class _AprilaireClientProtocol(asyncio.Protocol):
    """Protocol for interacting with the thermostat over socket connection"""

    def __init__(
        self,
        data_received_callback: Callable[[FunctionalDomain, int, dict[str, Any]], None],
        reconnect_action: Callable[[], None],
        logger: Logger,
    ) -> None:
        """Initialize the protocol"""
        self.data_received_callback = data_received_callback
        self.reconnect_action = reconnect_action
        self.logger = logger

        self.transport: asyncio.Transport = None

        self.command_queue = asyncio.Queue()

        self.sequence = 1

    async def _send_command(
        self,
        action: Action,
        functional_domain: FunctionalDomain,
        attribute: int,
        extra_payload: list[int] = None,
    ) -> None:
        """Send a command to the thermostat"""
        command_bytes = generate_command_bytes(
            self.sequence,
            action,
            functional_domain,
            attribute,
            extra_payload=extra_payload,
        )

        self.logger.debug(
            "Queuing data, sequence=%d, action=%s, functional_domain=%s, attribute=%d",
            self.sequence,
            str(action),
            str(functional_domain),
            attribute,
        )

        self.sequence = (self.sequence + 1) % 128

        await self.command_queue.put(command_bytes)

    def _empty_command_queue(self):
        try:
            for _ in range(self.command_queue.qsize()):
                self.command_queue.get_nowait()
                self.command_queue.task_done()
        except:  # pylint: disable=bare-except
            pass

    async def _process_command_queue(self):
        while True:
            command_bytes = await self.command_queue.get()

            if not command_bytes:
                continue

            if not self.transport:
                break

            self.logger.debug("Sending data %s", command_bytes.hex(" ", 1))

            self.transport.write(command_bytes)

    def connection_made(self, transport: asyncio.Transport):
        """Called when a connection has been made to the socket"""
        self.logger.info("Aprilaire connection made")

        self.transport = transport
        self._empty_command_queue()

        asyncio.ensure_future(self._process_command_queue())

        async def _update_status():
            await asyncio.sleep(2)

            await self.read_mac_address()
            await self.read_thermostat_status()
            await self.read_sensors()
            await self.configure_cos()
            await self.sync()

        asyncio.ensure_future(_update_status())

    def data_received(self, data: bytes) -> None:
        """Called when data has been received from the socket"""
        self.logger.info("Aprilaire data received")

        decoded_packets = decode_packet(data)

        for decoded_packet in decoded_packets:
            if "event" not in decoded_packet:
                self.logger.warning("Event data missing")
                return

            (action, functional_domain, attribute) = decoded_packet["event"]

            self.logger.debug(
                "Received data, action=%s, functional_domain=%s, attribute=%d",
                action,
                functional_domain,
                attribute,
            )

            if "error" in decoded_packet:
                error = decoded_packet["error"]

                if error != 0:
                    self.logger.error("Thermostat error: %d", error)

            if self.data_received_callback:
                asyncio.ensure_future(
                    self.data_received_callback(
                        functional_domain, attribute, decoded_packet
                    )
                )

    def connection_lost(self, exc: Exception | None) -> None:
        """Called when the connection to the socket has been lost"""
        self.logger.info("Aprilaire connection lost")

        if self.data_received_callback:
            asyncio.ensure_future(
                self.data_received_callback(
                    FunctionalDomain.NONE, 0, {"available": False}
                )
            )

        if self.reconnect_action:
            asyncio.ensure_future(self.reconnect_action())

    async def read_sensors(self):
        """Send a request for updated sensor data"""
        await self._send_command(Action.READ_REQUEST, FunctionalDomain.SENSORS, 2)

    async def read_control(self):
        """Send a request for updated control data"""
        await self._send_command(Action.READ_REQUEST, FunctionalDomain.CONTROL, 1)

    async def read_scheduling(self):
        """Send a request for updated scheduling data"""
        await self._send_command(Action.READ_REQUEST, FunctionalDomain.SCHEDULING, 4)

    async def update_mode(self, mode: int):
        """Send a request to update the mode"""
        await self._send_command(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            1,
            extra_payload=[
                mode,  # Mode
                0,  # Fan Mode (0 to not set)
                0,  # Heat Setpoint (0 to not set)
                0,  # Cool Setpoint (0 to not set)
            ],
        )

    async def update_fan_mode(self, fan_mode: int):
        """Send a request to update the fan mode"""
        await self._send_command(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            1,
            extra_payload=[
                0,  # Mode (0 to not set)
                fan_mode,  # Fan Mode
                0,  # Heat Setpoint (0 to not set)
                0,  # Cool Setpoint (0 to not set)
            ],
        )

    async def update_setpoint(self, cool_setpoint: int, heat_setpoint: int):
        """Send a request to update the setpoint"""
        await self._send_command(
            Action.WRITE,
            FunctionalDomain.CONTROL,
            1,
            extra_payload=[
                0,  # Mode
                0,  # Fan
                encode_temperature(heat_setpoint),  # Heat Setpoint
                encode_temperature(cool_setpoint),  # Cool Setpoint
            ],
        )

    async def set_hold(self, hold: int):
        """Send a request to set the hold status"""
        await self._send_command(
            Action.WRITE,
            FunctionalDomain.SCHEDULING,
            4,
            extra_payload=[
                hold,  # Hold
                0,  # Fan
                0,  # Heat Setpoint
                0,  # Cool Setpoint
                0,  # DEH Vacation
                0,  # End Minute
                0,  # End Hour
                0,  # End Date
                0,  # End Month
                0,  # End Year
            ],
        )

    async def sync(self):
        """Send a request to sync data"""
        await self._send_command(
            Action.WRITE,
            FunctionalDomain.STATUS,
            2,
            extra_payload=[
                1,  # Sync
            ],
        )

    async def configure_cos(self):
        """Send a request to configure the COS settings"""
        await self._send_command(
            Action.WRITE,
            FunctionalDomain.STATUS,
            1,
            extra_payload=[
                1,  # Installer Thermostat Settings
                0,  # Contractor Information
                0,  # Air Cleaning Installer Variable
                0,  # Humidity Control Installer Settings
                0,  # Fresh Air Installer Settings
                1,  # Thermostat Setpoint & Mode Settings
                0,  # Dehumidification Setpoint
                0,  # Humidification Setpoint
                0,  # Fresh Air Setting
                0,  # Air Cleaning Settings
                1,  # Thermostat IAQ Available
                0,  # Schedule Settings
                0,  # Away Settings
                0,  # Schedule Day
                1,  # Schedule Hold
                0,  # Heat Blast
                0,  # Service Reminders Status
                0,  # Alerts Status
                0,  # Alerts Settings
                0,  # Backlight Settings
                1,  # Thermostat Location & Name
                0,  # Reserved
                1,  # Controlling Sensor Values
                0,  # Over the air ODT update timeout
                1,  # Thermostat Status
                1,  # IAQ Status
                1,  # Model & Revision
                0,  # Support Module
                0,  # Lockouts
            ],
        )

    async def read_mac_address(self):
        """Send a request to get identification data (including MAC address)"""
        await self._send_command(
            Action.READ_REQUEST,
            FunctionalDomain.IDENTIFICATION,
            2,
        )

    async def read_thermostat_status(self):
        """Send a request for thermostat status"""
        await self._send_command(
            Action.READ_REQUEST,
            FunctionalDomain.CONTROL,
            7,
        )

    async def read_thermostat_name(self):
        """Send a reques for the thermostat name"""
        await self._send_command(
            Action.READ_REQUEST,
            FunctionalDomain.IDENTIFICATION,
            5,
        )


class AprilaireClient(SocketClient):
    """Client for sending/receiving data"""

    def __init__(
        self,
        host: str,
        port: int,
        data_received_callback: Callable[[dict[str, Any]], None],
        logger: Logger,
        reconnect_interval: int = None,
        retry_connection_interval: int = None,
    ) -> None:
        super().__init__(
            host,
            port,
            data_received_callback,
            logger,
            reconnect_interval,
            retry_connection_interval,
        )

        self.futures: dict[tuple[FunctionalDomain, int], list[asyncio.Future]] = {}

    def create_protocol(self):
        return _AprilaireClientProtocol(
            self.data_received, self._reconnect, self.logger
        )

    async def data_received(
        self, functional_domain: FunctionalDomain, attribute: int, data: dict[str, Any]
    ):
        """Called when data is received from the thermostat"""
        self.data_received_callback(data)

        if not functional_domain or not attribute:
            return

        future_key = (functional_domain, attribute)

        futures_to_complete = self.futures.pop(future_key, [])

        for future in futures_to_complete:
            try:
                future.set_result(data)
            except asyncio.exceptions.InvalidStateError:
                pass

    def state_changed(self):
        """Send data indicating the state as changed"""
        data = {
            "connected": self.connected,
            "stopped": self.stopped,
            "reconnecting": self.reconnecting,
        }

        self.data_received_callback(data)

    async def wait_for_response(
        self, functional_domain: FunctionalDomain, attribute: int, timeout: int = None
    ):
        """Wait for a response for a particular request"""

        loop = asyncio.get_event_loop()
        future = loop.create_future()

        future_key = (functional_domain, attribute)

        if future_key not in self.futures:
            self.futures[future_key] = []

        self.futures[future_key].append(future)

        try:
            return await asyncio.wait_for(future, timeout)
        except asyncio.exceptions.TimeoutError:
            self.logger.error(
                "Hit timeout of %d waiting for %s, %d",
                timeout,
                int(functional_domain),
                attribute,
            )
            return None

    async def read_sensors(self):
        """Send a request for updated sensor data"""
        await self.protocol.read_sensors()

    async def read_control(self):
        """Send a request for updated control data"""
        await self.protocol.read_control()

    async def read_scheduling(self):
        """Send a request for updated scheduling data"""
        await self.protocol.read_scheduling()

    async def update_mode(self, mode: int):
        """Send a request to update the mode"""
        await self.protocol.update_mode(mode)

    async def update_fan_mode(self, fan_mode: int):
        """Send a request to update the fan mode"""
        await self.protocol.update_fan_mode(fan_mode)

    async def update_setpoint(self, cool_setpoint: int, heat_setpoint: int):
        """Send a request to update the setpoint"""
        await self.protocol.update_setpoint(cool_setpoint, heat_setpoint)

    async def set_hold(self, hold: int):
        """Send a request to update the away status"""
        await self.protocol.set_hold(hold)

    async def sync(self):
        """Send a request to sync data"""
        await self.protocol.sync()

    async def read_mac_address(self):
        """Send a request to read the MAC address"""
        await self.protocol.read_mac_address()

    async def read_thermostat_name(self):
        """Send a request to read the thermostat name"""
        await self.protocol.read_thermostat_name()



"""Utilities for the Aprilaire integration"""

from __future__ import annotations

from crc import Calculator, Configuration
import math
from typing import Any

from .const import Action, FunctionalDomain

crc_calculator = Calculator(
    Configuration(
        width=8,
        polynomial=0x31,
        init_value=0,
        final_xor_value=0,
        reverse_input=False,
        reverse_output=False,
    )
)


def encode_temperature(temperature: float) -> int:
    """Encode a temperature value for sending to the thermostat"""
    is_negative = temperature < 0
    is_fraction = temperature % 1 >= 0.5

    return (
        math.floor(temperature)
        + (64 if is_fraction else 0)
        + (128 if is_negative else 0)
    )


def decode_temperature(raw_value: int) -> float:
    """Decode a temperature value from the thermostat"""
    temperature_value = float(int(raw_value & 63))

    raw_value = raw_value >> 6
    has_fraction = bool(raw_value & 1)
    if has_fraction:
        temperature_value += 0.5

    raw_value = raw_value >> 1
    is_positive = raw_value & 1 == 0
    if not is_positive:
        temperature_value = -temperature_value

    return temperature_value


def decode_humidity(raw_value: int) -> int:
    """Decode a humidity value from the thermostat"""
    if raw_value == 0 or raw_value >= 100:
        return None
    return raw_value


def generate_command_bytes(
    sequence: int,
    action: Action,
    functional_domain: FunctionalDomain,
    attribute: int,
    extra_payload: list[int] = None,
) -> list[int]:
    """Generate the data to send to the thermostat"""
    payload = [int(action), int(functional_domain), attribute]
    if extra_payload:
        payload.extend(extra_payload)
    (payload_length_high, payload_length_low) = _encode_int_value(len(payload))
    result = [1, sequence, payload_length_high, payload_length_low]
    result.extend(payload)
    result.append(generate_crc(result))
    return bytes(result)


def pad_list(lst: list[Any], length: int, pad: Any = 0):
    """Pad a list to a minimum length"""
    return lst + [pad] * (length - len(lst))


def generate_crc(lst: list[int]):
    """Generate a CRC checksum"""
    return crc_calculator.checksum(bytes(lst))


def verify_crc(lst: list[int], crc: int):
    """Verify a CRC checksum"""
    return crc_calculator.verify(bytes(lst), crc)


def _encode_int_value(value: int):
    return ((value >> 8) & 0xFF, value & 0xFF)



"""Functions for handling response data from the thermostat"""

from __future__ import annotations

import logging

from enum import Enum
from typing import Any

from .const import Action, FunctionalDomain
from .utils import decode_humidity, decode_temperature


class ValueType(Enum):
    """Parsable value types from data"""

    INTEGER = 1
    INTEGER_REQUIRED = 2
    TEMPERATURE = 3
    TEMPERATURE_REQUIRED = 4
    HUMIDITY = 5
    MAC_ADDRESS = 6
    TEXT = 7


MAPPING = {
    Action.READ_RESPONSE: {
        FunctionalDomain.SETUP: {
            1: [
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                ("away_available", ValueType.INTEGER),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
            ]
        },
        FunctionalDomain.CONTROL: {
            1: [
                ("mode", ValueType.INTEGER_REQUIRED),
                ("fan_mode", ValueType.INTEGER_REQUIRED),
                ("heat_setpoint", ValueType.TEMPERATURE_REQUIRED),
                ("cool_setpoint", ValueType.TEMPERATURE_REQUIRED),
            ],
            7: [
                ("thermostat_modes", ValueType.INTEGER),
                ("air_cleaning_available", ValueType.INTEGER),
                ("ventilation_available", ValueType.INTEGER),
                ("dehumidification_available", ValueType.INTEGER),
                ("humidification_available", ValueType.INTEGER),
            ],
        },
        FunctionalDomain.SCHEDULING: {
            4: [
                ("hold", ValueType.INTEGER),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
                (None, None),
            ],
        },
        FunctionalDomain.SENSORS: {
            1: [
                ("built_in_temperature_sensor_status", ValueType.INTEGER),
                ("built_in_temperature_sensor_value", ValueType.TEMPERATURE),
                ("wired_remote_temperature_sensor_status", ValueType.INTEGER),
                ("wired_remote_temperature_sensor_value", ValueType.TEMPERATURE),
                ("wired_outdoor_temperature_sensor_status", ValueType.INTEGER),
                ("wired_outdoor_temperature_sensor_value", ValueType.TEMPERATURE),
                ("built_in_humidity_sensor_status", ValueType.INTEGER),
                ("built_in_humidity_sensor_value", ValueType.HUMIDITY),
                ("rat_sensor_status", ValueType.INTEGER),
                ("rat_sensor_value", ValueType.TEMPERATURE),
                ("lat_sensor_status", ValueType.INTEGER),
                ("lat_sensor_value", ValueType.TEMPERATURE),
                ("wireless_outdoor_temperature_sensor_status", ValueType.INTEGER),
                ("wireless_outdoor_temperature_sensor_value", ValueType.TEMPERATURE),
                ("wireless_outdoor_humidity_sensor_status", ValueType.INTEGER),
                ("wireless_outdoor_humidity_sensor_value", ValueType.HUMIDITY),
            ],
            2: [
                ("indoor_temperature_controlling_sensor_status", ValueType.INTEGER),
                ("indoor_temperature_controlling_sensor_value", ValueType.TEMPERATURE),
                ("outdoor_temperature_controlling_sensor_status", ValueType.INTEGER),
                ("outdoor_temperature_controlling_sensor_value", ValueType.TEMPERATURE),
                ("indoor_humidity_controlling_sensor_status", ValueType.INTEGER),
                ("indoor_humidity_controlling_sensor_value", ValueType.HUMIDITY),
                ("outdoor_humidity_controlling_sensor_status", ValueType.INTEGER),
                ("outdoor_humidity_controlling_sensor_value", ValueType.HUMIDITY),
            ],
        },
        FunctionalDomain.STATUS: {
            2: [
                ("synced", ValueType.INTEGER),
            ],
            6: [
                ("heating_equipment_status", ValueType.INTEGER),
                ("cooling_equipment_status", ValueType.INTEGER),
                ("progressive_recovery", ValueType.INTEGER),
                ("fan_status", ValueType.INTEGER),
            ],
            7: [
                ("dehumidification_status", ValueType.INTEGER),
                ("humidification_status", ValueType.INTEGER),
                ("ventilation_status", ValueType.INTEGER),
                ("air_cleaning_status", ValueType.INTEGER),
            ],
            8: [
                ("error", ValueType.INTEGER),
            ],
        },
        FunctionalDomain.IDENTIFICATION: {
            1: [
                ("hardware_revision", ValueType.INTEGER),
                ("firmware_major_revision", ValueType.INTEGER),
                ("firmware_minor_revision", ValueType.INTEGER),
                ("protocol_major_revision", ValueType.INTEGER),
                ("model_number", ValueType.INTEGER),
                ("gainspan_firmware_major_revision", ValueType.INTEGER),
                ("gainspan_firmware_minor_revision", ValueType.INTEGER),
            ],
            2: [
                ("mac_address", ValueType.MAC_ADDRESS),
            ],
            4: [
                ("location", ValueType.TEXT, 7),
                ("name", ValueType.TEXT, 15),
            ],
        },
    }
}

MAPPING[Action.COS] = MAPPING[Action.READ_RESPONSE]
MAPPING[Action.WRITE] = MAPPING[Action.READ_RESPONSE]

LOG_NAME = ""
_LOGGER = logging.getLogger(LOG_NAME)


def decode_packet(data: bytes) -> list[dict[str, Any]]:
    """Decode the response data from the thermostat"""

    (action, functional_domain, attribute) = decode_packet_header(data)

    if action == Action.NACK:
        _LOGGER.warning("NACK received: %s", data.hex(" ", 1))
        return []

    if (
        action not in MAPPING
        or functional_domain not in MAPPING[action]
        or attribute not in MAPPING[action][functional_domain]
    ):
        _LOGGER.debug(
            "Unhandled command, action=%s, functional_domain=%s, attribute=%d, data=%s",
            str(action),
            str(functional_domain),
            attribute,
            data.hex(" ", 1),
        )

        return []

    _LOGGER.debug("Reading data=%s", data.hex(" ", 1))

    results: list[dict[str, Any]] = []

    current_result = {"event": (action, functional_domain, attribute)}

    i = 0

    while i < len(data):
        if i == 0:
            current_result["revision"] = data[i]
        elif i == 1:
            current_result["sequence"] = data[i]
        elif i == 2:
            current_result["count"] = data[2] << 2 | data[3]
            i += 1
        elif i == 4:
            j = 0

            attribute_index = 0

            while j < current_result["count"]:
                if j < 3:
                    pass
                elif attribute_index >= len(
                    MAPPING[action][functional_domain][attribute]
                ):
                    pass
                else:
                    attribute_info = MAPPING[action][functional_domain][attribute][
                        attribute_index
                    ]

                    (attribute_name, value_type, extra_attribute_info) = (
                        attribute_info[0],
                        attribute_info[1],
                        attribute_info[2:],
                    )

                    if attribute_name is None or value_type is None:
                        j += 1
                        attribute_index += 1
                        continue

                    data_value = data[i + j]

                    if value_type == ValueType.INTEGER:
                        current_result[attribute_name] = data_value
                    elif value_type == ValueType.INTEGER_REQUIRED:
                        if data_value is not None and data_value != 0:
                            current_result[attribute_name] = data_value
                    elif value_type == ValueType.HUMIDITY:
                        current_result[attribute_name] = decode_humidity(data_value)
                    elif value_type == ValueType.TEMPERATURE:
                        current_result[attribute_name] = decode_temperature(data_value)
                    elif value_type == ValueType.TEMPERATURE_REQUIRED:
                        if data_value is not None and data_value != 0:
                            current_result[attribute_name] = decode_temperature(
                                data_value
                            )
                    elif value_type == ValueType.MAC_ADDRESS:
                        mac_address_components = []

                        for _ in range(0, 6):
                            mac_address_components.append(f"{data[i + j]:x}")
                            j += 1

                        current_result[attribute_name] = ":".join(
                            mac_address_components
                        )
                    elif value_type == ValueType.TEXT:
                        text_length = extra_attribute_info[0]

                        text = ""

                        for _ in range(0, text_length):
                            current_value = (
                                " " if data[i + j] == 0 else chr(data[i + j])
                            )
                            text += current_value
                            j += 1

                        text = text.strip(" ")

                        current_result[attribute_name] = text

                    attribute_index += 1
                j += 1
            i += j
        else:
            results.extend(decode_packet(data[i:]))
            break

        i += 1

    results.insert(0, current_result)

    return results


def decode_packet_header(data):
    """Read the header from a packet"""
    try:
        action = Action(int(data[4]))
        functional_domain = FunctionalDomain(int(data[5]))
        attribute = int(data[6])
    except ValueError:
        return (Action.NONE, FunctionalDomain.NONE, 0)

    return (action, functional_domain, attribute)



"""Mock server for testing Aprilaire integration"""

from __future__ import annotations

import argparse
import asyncio
from asyncio import Queue
import logging

from .const import Action, FunctionalDomain
from .packet import decode_packet, decode_packet_header
from .utils import encode_temperature, generate_command_bytes, pad_list

COS_FREQUENCY = 30
QUEUE_FREQUENCY = 0.5


class CustomFormatter(logging.Formatter):
    """Custom logging formatter"""

    green = "\x1b[32;20m"
    cyan = "\x1b[36;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    log_format = "%(asctime)s %(levelname)s [%(name)s] %(message)s"

    FORMATS = {
        logging.DEBUG: cyan + log_format + reset,
        logging.INFO: green + log_format + reset,
        logging.WARNING: yellow + log_format + reset,
        logging.ERROR: red + log_format + reset,
        logging.CRITICAL: bold_red + log_format + reset,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


_LOGGER = logging.getLogger("aprilaire.mock_server")
_LOGGER.setLevel(logging.DEBUG)

ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

ch.setFormatter(CustomFormatter())

_LOGGER.addHandler(ch)


class _AprilaireServerProtocol(asyncio.Protocol):
    def __init__(self):
        self.transport: asyncio.Transport = None

        self.mode = 5
        self.fan_mode = 2
        self.cool_setpoint = 25
        self.heat_setpoint = 20
        self.hold = 0

        self.name = "ttt12347"
        self.location = "1234"

        self.queue = Queue()

        self.sequence = 1

    def _generate_thermostat_status_command_bytes(self):
        heating_equipment_status = {2: 2, 4: 7}.get(self.mode, 0)

        cooling_equipment_status = {3: 2, 5: 2}.get(self.mode, 0)

        fan_status = 1 if self.fan_mode == 1 or self.fan_mode == 2 else 0

        return generate_command_bytes(
            self.sequence + 127,
            Action.COS,
            FunctionalDomain.STATUS,
            6,
            [heating_equipment_status, cooling_equipment_status, 0, fan_status],
        )

    async def send_status(self):
        """Send the current status"""

        await self.queue.put(
            generate_command_bytes(
                self.sequence + 127,
                Action.READ_RESPONSE,
                FunctionalDomain.IDENTIFICATION,
                2,
                [1, 2, 3, 4, 5, 6],
            )
        )

        self.sequence = (self.sequence + 1) % 128

        await self.queue.put(
            generate_command_bytes(
                self.sequence + 127,
                Action.COS,
                FunctionalDomain.CONTROL,
                1,
                [
                    self.mode,
                    self.fan_mode,
                    encode_temperature(self.heat_setpoint),
                    encode_temperature(self.cool_setpoint),
                ],
            )
            + generate_command_bytes(
                self.sequence + 127,
                Action.COS,
                FunctionalDomain.SENSORS,
                2,
                [0, encode_temperature(25), 0, encode_temperature(20), 0, 50, 0, 40],
            )
            + generate_command_bytes(
                self.sequence + 127, Action.COS, FunctionalDomain.STATUS, 2, [1]
            )
            + generate_command_bytes(
                self.sequence + 127,
                Action.COS,
                FunctionalDomain.STATUS,
                7,
                [2, 2, 2, 2],
            )
            + generate_command_bytes(
                self.sequence + 127,
                Action.COS,
                FunctionalDomain.CONTROL,
                7,
                [6, 1, 1, 1],
            )
            + generate_command_bytes(
                self.sequence + 127,
                Action.COS,
                FunctionalDomain.SETUP,
                1,
                list([0] * 26) + [1] + list([0] * 17),
            )
            + generate_command_bytes(
                self.sequence + 127,
                Action.COS,
                FunctionalDomain.SCHEDULING,
                4,
                [self.hold] + list([0] * 9),
            )
            + generate_command_bytes(
                self.sequence + 127,
                Action.COS,
                FunctionalDomain.IDENTIFICATION,
                1,
                [66, 10, 2, 15, 1, 14, 3],
            )
            + generate_command_bytes(
                self.sequence + 127,
                Action.READ_RESPONSE,
                FunctionalDomain.IDENTIFICATION,
                4,
                (
                    pad_list([ord(c) for c in self.location], 8, 0)
                    + pad_list([ord(c) for c in self.name], 16, 0)
                ),
            )
            + self._generate_thermostat_status_command_bytes()
        )

        self.sequence = (self.sequence + 1) % 128

    async def cos_loop(self):
        """Send the current status (COS) periodically"""
        await asyncio.sleep(2)

        while self.transport:
            await self.send_status()
            await asyncio.sleep(COS_FREQUENCY)

    async def queue_loop(self):
        """Periodically send items from the queue"""
        while self.transport:
            command_bytes = await self.queue.get()

            if self.transport:
                _LOGGER.info("Sent data: %s", command_bytes.hex(" ", 1))

                self.transport.write(command_bytes)

            await asyncio.sleep(QUEUE_FREQUENCY)

    def connection_made(self, transport):
        _LOGGER.info("Connection made")

        self.transport = transport

        asyncio.ensure_future(self.cos_loop())
        asyncio.ensure_future(self.queue_loop())

    def data_received(self, data: bytes) -> None:
        _LOGGER.info("Received data: %s", data.hex(" ", 1))

        (action, functional_domain, attribute) = decode_packet_header(data)

        if action == Action.READ_REQUEST:
            if functional_domain == FunctionalDomain.SENSORS:
                if attribute == 2:
                    self.queue.put_nowait(
                        generate_command_bytes(
                            self.sequence + 127,
                            Action.READ_RESPONSE,
                            FunctionalDomain.SENSORS,
                            2,
                            [
                                0,
                                encode_temperature(22),
                                0,
                                encode_temperature(10),
                                0,
                                50,
                                0,
                                40,
                            ],
                        )
                    )

                    self.sequence = (self.sequence + 1) % 128
            elif functional_domain == FunctionalDomain.SCHEDULING:
                if attribute == 4:
                    self.queue.put_nowait(
                        generate_command_bytes(
                            self.sequence + 127,
                            Action.COS,
                            FunctionalDomain.SCHEDULING,
                            4,
                            [self.hold] + list([0] * 9),
                        )
                    )

                    self.sequence = (self.sequence + 1) % 128
            elif functional_domain == FunctionalDomain.IDENTIFICATION:
                if attribute == 4:
                    self.queue.put_nowait(
                        generate_command_bytes(
                            self.sequence + 127,
                            Action.READ_RESPONSE,
                            FunctionalDomain.IDENTIFICATION,
                            4,
                            (
                                pad_list([ord(c) for c in self.location], 8, 0)
                                + pad_list([ord(c) for c in self.name], 16, 0)
                            ),
                        )
                    )

                    self.sequence = (self.sequence + 1) % 128
        elif action == Action.WRITE:
            if functional_domain == FunctionalDomain.CONTROL:
                if attribute == 1:
                    decoded_packets = decode_packet(data)

                    for decoded_packet in decoded_packets:
                        if "mode" in decoded_packet:
                            new_mode = decoded_packet["mode"]

                            if new_mode != 0:
                                self.mode = new_mode
                                self.hold = 0

                        if "fan_mode" in decoded_packet:
                            new_fan_mode = decoded_packet["fan_mode"]

                            if new_fan_mode != 0:
                                self.fan_mode = new_fan_mode

                        if "heat_setpoint" in decoded_packet:
                            new_heat_setpoint = decoded_packet["heat_setpoint"]

                            if new_heat_setpoint != 0:
                                self.heat_setpoint = new_heat_setpoint
                                self.hold = 1

                        if "cool_setpoint" in decoded_packet:
                            new_cool_setpoint = decoded_packet["cool_setpoint"]

                            if new_cool_setpoint != 0:
                                self.cool_setpoint = new_cool_setpoint
                                self.hold = 1

                        self.queue.put_nowait(
                            generate_command_bytes(
                                self.sequence + 127,
                                Action.COS,
                                FunctionalDomain.CONTROL,
                                1,
                                [
                                    self.mode,
                                    self.fan_mode,
                                    encode_temperature(self.heat_setpoint),
                                    encode_temperature(self.cool_setpoint),
                                ],
                            )
                        )

                        self.sequence = (self.sequence + 1) % 128

                        self.queue.put_nowait(
                            self._generate_thermostat_status_command_bytes()
                        )

                        self.queue.put_nowait(
                            generate_command_bytes(
                                self.sequence + 127,
                                Action.COS,
                                FunctionalDomain.SCHEDULING,
                                4,
                                [self.hold] + list([0] * 9),
                            )
                        )

                        self.sequence = (self.sequence + 1) % 128
            elif functional_domain == FunctionalDomain.SCHEDULING:
                if attribute == 4:
                    decoded_packets = decode_packet(data)

                    for decoded_packet in decoded_packets:
                        if "hold" in decoded_packet:
                            self.hold = decoded_packet["hold"]

                    self.queue.put_nowait(
                        generate_command_bytes(
                            self.sequence + 127,
                            Action.COS,
                            FunctionalDomain.SCHEDULING,
                            4,
                            [self.hold] + list([0] * 9),
                        )
                    )

                    self.sequence = (self.sequence + 1) % 128
            elif functional_domain == FunctionalDomain.STATUS:
                if attribute == 2:
                    asyncio.ensure_future(self.send_status())

    def connection_lost(self, exc: Exception | None) -> None:
        _LOGGER.info("Connection lost")
        self.transport = None


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-H", "--host", default="localhost")
    parser.add_argument("-p", "--port", default=7001)

    args = parser.parse_args()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    coro = loop.create_server(_AprilaireServerProtocol, args.host, args.port)

    server = loop.run_until_complete(coro)

    _LOGGER.info("Server listening on %s port %d", args.host, args.port)

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        server.close()
        loop.close()
